//'WWW''''''''''''''''''''''''''''''WW''''WW''''''''WW''''''''''WWW'''''''
//WW  W                             WW              WW           WW       
//WW     WWW WWWW W   WW  W WW  WWW WW    WW  WWWW WWWW     WWW  WW   WWW 
//WW    W WW WW WW W WWWW WW   W WW WW    WW WWW    WW     WW W  WW  W WW 
//WW  W W WW WW WW W WW   WW   W WW WW    WW   WWW  WW     WW    WW  W WW 
// WWW   WWW WW WW W  WWW WW    WWW WWWWW WW WWWW   WW      WWW WWWW  WWW 
//                                                                        
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

#include "CameraList.h"

#include "Main.h"

#include "debug.h"

// NOTE:
//
// the following tridiagonal matrix solver and cubic interpolation code are
// translated from a java program written by T. von Petersdorff
// 
//        http://www.wam.umd.edu/~petersd/interp.html

  
// solve linear system with tridiagonal n by n matrix a
// using Gaussian elimination *without* pivoting
// where   a(i,i-1) = sub[i]  for 2<=i<=n
//         a(i,i)   = diag[i] for 1<=i<=n
//         a(i,i+1) = sup[i]  for 1<=i<=n-1
// (the values sub[1], sup[n] are ignored)
//  right hand side vector b[1:n] is overwritten with solution 
//  NOTE: 1...n is used in all arrays, 0 is unused */
void CameraList::solveTridiag(float sub[], float diag[], float sup[],
			      float b[], int n){
    
  int i;
  // factorization and forward substitution
  for(i=2; i<=n; i++){
    sub[i] = sub[i]/diag[i-1];
    diag[i] = diag[i] - sub[i]*sup[i-1];
    b[i] = b[i] - sub[i]*b[i-1];
  }
  b[n] = b[n]/diag[n];
  for(i=n-1;i>=1;i--){
    b[i] = (b[i] - sup[i]*b[i+1])/diag[i];
  }
}       
  
// x[i] are assumed evenly spaced from 0..1
// y[i] are from 0..np-1
float CameraList::cubicInterp(float y[], int np, float t){
//  return 1.0;

  if (np<0){
    cerr << "Wrong usage of " << __FUNCTION__ << endl;
    exit(1);
  }
  else if (np==0)
    return 0.0;
  else if (np==1)
    return y[0];

  // np>1
  float yi;
  float a[np];
  float h[np];        // distance between points along x axis
  float t1;           // fraction of h[i] from first point
  float t2;           // fraction of h[i] from second point

  a[0] = 0.0;    
  for (int i=1; i<=np-1; i++)
    h[i] = 1.0/(np-1); // x[i] - x[i-1] : evenly spaced, in our case.

  if (np>2){
    float sub[np-1];
    float diag[np-1];
    float sup[np-1];

    a[np-1] = 0.0;

    // initialize tridiagonal matrix
    for (int i=1; i<=np-2; i++){
      diag[i] = (h[i] + h[i+1])/3.0;
      sup[i] = h[i+1]/6.0;
      sub[i] = h[i]/6.0;
      a[i] = (y[i+1]-y[i])/h[i+1]-(y[i]-y[i-1])/h[i];
    }
	
    solveTridiag(sub,diag,sup,a,np-2);
  }
      
  int i;
  float j;
  i = (int)( (np-1)*t ) + 1;    // interval
  if (i > np-1) i = np-1;
  j = t*(np-1) - (i-1);         // percent along interval
      
  t1 = h[i]*j;
  t2 = h[i] - t1;
      
  yi =  (-a[i-1]/6.0*(t2+h[i])*t1 + y[i-1]) * t2;
  yi += (-a[i]  /6.0*(t1+h[i])*t2 + y[i])   * t1;
  yi /= h[i];
  
  return yi;
}


int CameraList::readList(string camFile){
  struct stat file_stat;
  
  if (camFile.empty())
    camFile = "camlist.iv";

  // test for presence of file
  if (stat(camFile.c_str(), &file_stat) == 0){

    // open the file
    SoInput *camIn = new SoInput();
    if (camIn->openFile("camlist.iv")){

      // read contents of file
      SoSeparator *camScene = SoDB::readAll(camIn);
      camScene->ref();

      // cam scene is assumed to be only SoPerspectiveCameras
      int numCams = camScene->getNumChildren();
      cerr << "Adding " << numCams << " cameras to camera list.\n";

      // loop through cameras
      for (int i=0; i<numCams; i++){
	SoNode *node = camScene->getChild(i);

	// add cameras to camList
	if (node->isOfType(SoPerspectiveCamera::getClassTypeId()))
	  addCamera((SoPerspectiveCamera *)node);
      }

      // delete list
      camScene->unref();
    }

    // delete SoInput
    delete camIn;

    return(1);
  }
  else {
    cerr << "No camera list file camlist.iv found. Not using any camera paths.\n";
    return(0);
  }
}


void CameraList::positionCameraLinear(SoPerspectiveCamera *cam, float time){

  int numcams = camlist.size();
  if (numcams<1)
    return;
    
  int first, second;
  first = (int)( (numcams-1)*time );
  second = MIN2(first+1, numcams-1);

  // adjust time
  if (numcams==1){
    time = 0.0;
    second = first;
  }
  else
    time = (time-first*1.0/(numcams-1)) / (1.0/(numcams-1));

  SoPerspectiveCamera *cam1 = camlist[first];
  SoPerspectiveCamera *cam2 = camlist[second];

  // interpolate values

  // position
  SbVec3f pos1 = cam1->position.getValue();
  SbVec3f pos2 = cam2->position.getValue();
  cam->position.setValue( LERP(pos1, pos2, time) );

  // orientation
  // interpolate quaternions for smooth rotation
//  const float *rot1 = cam1->orientation.getValue().getValue();
//  const float *rot2 = cam2->orientation.getValue().getValue();
//  float rot3[4];

//  float cosom = (rot1[0]*rot2[0]+
// 		rot1[1]*rot2[1]+
// 		rot1[2]*rot2[2]+
// 		rot1[3]*rot2[3]);
 
//  if (cosom<0.0){
//    cosom = -cosom;
//    rot3[0] = -rot2[0];
//    rot3[1] = -rot2[1];
//    rot3[2] = -rot2[2];
//    rot3[3] = -rot2[3];
//  }
//  else {
//    rot3[0] = rot2[0];
//    rot3[1] = rot2[1];
//    rot3[2] = rot2[2];
//    rot3[3] = rot2[3];
//  }

//  float om = acos(cosom);
//  float sinom = sin(om);
//  float sp = sin((1-time)*om)/sinom;
//  float sq = sin(time*om)/sinom;
 
//   cam->orientation.setValue(
//     rot1[0]*sp + rot3[0]*sq,
//     rot1[1]*sp + rot3[1]*sq,
//     rot1[2]*sp + rot3[2]*sq,
//     rot1[3]*sp + rot3[3]*sq
//     );

  cam->orientation.setValue(SbRotation::slerp(cam1->orientation.getValue(),
					      cam2->orientation.getValue(), time));



  // focalDistance
  float fd1 = cam1->focalDistance.getValue();
  float fd2 = cam2->focalDistance.getValue();
  cam->focalDistance.setValue( LERP(fd1, fd2, time) );

  // heightAngle
  float ha1 = cam1->heightAngle.getValue();
  float ha2 = cam2->heightAngle.getValue();
  cam->heightAngle.setValue( LERP(ha1, ha2, time) );

    
}

// quaternion multiplication algorithm
// from 3D Math Primer for Graphics and Game Development
// by Fletcher Dunn and Ian Parberry
inline SbVec4f quadMult(SbVec4f a, SbVec4f b){
  float w1, w2;
  SbVec3f v1, v2;
  a.getValue(w1, v1[0], v1[1], v1[2]);
  b.getValue(w2, v2[0], v2[1], v2[2]);
  SbVec3f c = w1*v2 + w2*v1 + v1.cross(v2);
  return SbVec4f(w1*w2-v1.dot(v2),
		 c[0], c[1], c[2]);
}

inline SbVec4f quadInv(SbVec4f x){
  return( SbVec4f(x[0], -x[1], -x[2], -x[3])/x.length() );
}

inline SbVec4f log(SbVec4f x){
  if (x[0]>=1.0)
    return SbVec4f(0,0,0,0);
  float alpha = acos(x[0]);
  float b = alpha/sin(alpha);
  return SbVec4f(0, b*x[1], b*x[2], b*x[3]);
}

inline SbVec4f exp(SbVec4f x){
  SbVec3f n(x[1], x[2], x[3]);
  float alpha = n.length();
  if (alpha<=0.0)
    return SbVec4f(cos(alpha), 0.0, 0.0, 0.0);
  SbVec3f c = sin(alpha)*n/alpha;
  return SbVec4f( cos(alpha), c[0], c[1], c[2]);
}

SbRotation squad(SbVec4f q[4], float h){
  SbVec4f s[4];
  for (int i=1; i<=2; i++){
    
    s[i] =
      quadMult(exp( -(log(quadMult(q[i+1], quadInv(q[i]))) +
		      log(quadMult(q[i-1], quadInv(q[i]))) )/4.0 ), q[i]); 
      
  }

  SbRotation qi(q[1].getValue());
  SbRotation q1(q[2].getValue());
  SbRotation si(s[1].getValue());
  SbRotation s1(s[2].getValue());

  SbRotation a = SbRotation::slerp(qi,q1,h);
  SbRotation b = SbRotation::slerp(si,s1,h);
  
  return SbRotation::slerp(a, b, 2*h*(1.0-h));
}

void CameraList::positionCameraCubic(SoPerspectiveCamera *cam, float time){

  int numcams = camlist.size();
  if (numcams<1)
    return;

  // do linear interpolation if only two cams
  if (numcams<=2){
    positionCameraLinear(cam, time);
    return;
  }
    
  // interpolate values
  SbVec3f pos, newPos = SbVec3f(0.0, 400.0, -800.0);
  SoPerspectiveCamera *cami;
  float x[numcams];
  float interp;
    
  // position
  for (int c=0; c<3; c++){
      
    // populate x[]
    for (int i=0; i<numcams; i++){
      cami = camlist[i];
      pos = cami->position.getValue();
      x[i] = pos[c];
    }

    // get interpolated value
    interp = cubicInterp(x, numcams, time);
    newPos[c] = interp;
  }
  cam->position.setValue( newPos );

  /*
  // debugging
  static int pnts = 0;
  if (pnts++<200)
    showPoint(newPos, 5.0);
  */

  // orientation
  // interpolate quaternions for smooth rotation

  int k = (int)(time*(numcams-1));
  float w = 1.0/(numcams-1);
  float s = (time-k*w)/w;

  if (k==numcams-1){
    cam->orientation = camlist[k]->orientation.getValue();
  }
  if (k==0 || k==numcams-2){
    SbRotation r1(camlist[k  ]->orientation.getValue());
    SbRotation r2(camlist[k+1]->orientation.getValue());
    cam->orientation.setValue(SbRotation::slerp(r1, r2, s));
  }
  else {
    SbVec4f q[4];
    for (int i=0; i<4; i++){
      int j = k-1+i;
      if (j<0) j = 0;
      if (j>=numcams) j = numcams-1;
      q[i].setValue(camlist[j]->orientation.getValue().getValue());

      // pick closest quaternion
      if (i>0 && q[i].dot(q[i-1])<0.0)
	q[i] = -q[i];
    }
    cam->orientation = squad(q,s);
  }
    
//  float newRot[4];
//   for (int c=0; c<4; c++){      
//     // populate x[]
//     for (int i=0; i<numcams; i++){
//       cami = camlist[i];
//       x[i] = cami->orientation.getValue().getValue()[c];
//     }

//     // get interpolated value
//     interp = cubicInterp(x, numcams, time);
//     newRot[c] = interp;
//   }
//  cam->orientation.setValue( newRot[0], newRot[1], newRot[2], newRot[3] );

  // focalDistance
  // populate x[]
  for (int i=0; i<numcams; i++){
    cami = camlist[i];
    x[i] = cami->focalDistance.getValue();
  }
    
  // get interpolated value
  interp = cubicInterp(x, numcams, time);
  cam->focalDistance.setValue( interp );

  // heightAngle
  // populate x[]
  for (int i=0; i<numcams; i++){
    cami = camlist[i];
    x[i] = cami->heightAngle.getValue();
  }
    
  // get interpolated value
  interp = cubicInterp(x, numcams, time);
  cam->heightAngle.setValue( interp );
    
}
