#ifndef JULIA_H
#define JULIA_H

#include <iostream>
#include <iomanip>
#include "Primitive.h"
#include "Material.h"

using namespace std;


class Quat {
public:
  double r,i,j,k;
  explicit Quat(){};
  explicit Quat(double r_, double i_, double j_, double k_):r(r_),i(i_),j(j_),k(k_){}
  Quat operator+(const Quat &b) const {
    return Quat(r+b.r, i+b.i, j+b.j, k+b.k);
  }
  Quat operator-(const Quat &b) const {
    return Quat(r-b.r, i-b.i, j-b.j, k-b.k);
  }
  Quat operator*(double c) const {
    return Quat(r*c, i*c, j*c, k*c);
  }
  friend Quat operator*(double c, const Quat &a);
  friend ostream& operator<< (ostream& os, const Quat& x);
    
  Quat operator*(const Quat &b) const {
    // i*i = j*j = k*k = i*j*k = -1
    //
    // i*j = k    j*i = -k
    // j*k = i    k*j = -i
    // k*i = j    i*k = -j
    //
    // (a0 + a1*i + a2*j + a3*k) * (b0 + b1*i + b2*j + b3*k) =
    //
    // (a0 + a1*i + a2*j + a3*k)*b0 +
    // (a0 + a1*i + a2*j + a3*k)*b1*i +
    // (a0 + a1*i + a2*j + a3*k)*b2*j +
    // (a0 + a1*i + a2*j + a3*k)*b3*k =
    //
    // (a0*b0   + a1*b0*i   + a2*b0*j   + a3*b0*k) +
    // (a0*b1*i + a1*b1*i*i + a2*b1*j*i + a3*b1*k*i) +
    // (a0*b2*j + a1*b2*i*j + a2*b2*j*j + a3*b2*k*j) +
    // (a0*b3*k + a1*b3*i*k + a2*b3*j*k + a3*b3*k*k) =
    //
    // (a0*b0   + a1*b0*i + a2*b0*j + a3*b0*k) +
    // (a0*b1*i - a1*b1   - a2*b1*k + a3*b1*j) +
    // (a0*b2*j + a1*b2*k - a2*b2   - a3*b2*i) +
    // (a0*b3*k - a1*b3*j + a2*b3*i - a3*b3) =
    //
    // ( a0*b0 + a1*b0*i + a2*b0*j + a3*b0*k) +
    // (-a1*b1 + a0*b1*i + a3*b1*j - a2*b1*k) +
    // (-a2*b2 - a3*b2*i + a0*b2*j + a1*b2*k) +
    // (-a3*b3 + a2*b3*i - a1*b3*j + a0*b3*k) =
    //
    // Subtituting
    //   a0=a.r a1=a.i a2=a.j a3=a.k
    //   b0=b.r b1=b.i b2=b.j b3=b.k
    //
    // ( a.r*b.r + a.i*b.r*i + a.j*b.r*j + a.k*b.r*k) +
    // (-a.i*b.i + a.r*b.i*i + a.k*b.i*j - a.j*b.i*k) +
    // (-a.j*b.j - a.k*b.j*i + a.r*b.j*j + a.i*b.j*k) +
    // (-a.k*b.k + a.j*b.k*i - a.i*b.k*j + a.r*b.k*k) =

    return Quat(
//       r*b.r - i*b.i - j*b.j - k*b.k,
//       i*b.r + r*b.i - k*b.j + j*b.k,
//       j*b.r + k*b.i + r*b.j - i*b.k,
//       k*b.r - j*b.i + i*b.j + r*b.k);

      (r*b.r - i*b.i) - (j*b.j + k*b.k),
      (i*b.r + r*b.i) + (j*b.k - k*b.j),
      (j*b.r + k*b.i) + (r*b.j - i*b.k),
      (k*b.r - j*b.i) + (i*b.j + r*b.k));
  }

  double dot(const Quat &q) const { return r*q.r + i*q.i + j*q.j + k*q.k; }
  double len2() const { return r*r + i*i + j*j + k*k; }
  double len() const { return sqrt(len2()); }
  Quat norm() const { return (*this)*(1/len()); }
  Quat operator*(double c) { return Quat(r*c, i*c, j*c, k*c); }
};

inline Quat operator*(double c, const Quat &q) {
  return Quat(q.r*c, q.i*c, q.j*c, q.k*c);
}

inline ostream& operator<< (ostream& os, const Quat& x){
  os << setiosflags(ios::showpoint | ios::fixed) << setprecision(8);
  os << "(" << x.r << ", " << x.i << ", " << x.j << ", " << x.k << ")";

  //resetios(os);
  const int defaultPrc = 6;
  os << resetiosflags(ios::floatfield) << setprecision(defaultPrc);
  
  return os;
}

class SimpleSphere {
public:
  double rad, rad2;     // radius, radius^2
  SbVec3f pos;          // position
  SimpleSphere(){}
  SimpleSphere(float rad_, SbVec3f pos_) : rad(rad_), pos(pos_) {
    rad2 = rad*rad;
  }
  bool intersect(const Ray &r, double t0, double &min, double &max){
    SbVec3f op = pos-r.pos;   // solve t^2*d.d + 2*t*(o-p).d + (o-p).(o-p)-R^2 = 0
    double b = op.dot(r.dir);
    double det = b*b - op.dot(op) + rad2;
    if (det<0)
      return false;
    det = sqrt(det);
    min = b-det;
    max = b+det;
    return min>t0;
  }
};

class Julia : public Primitive {

private:

  static Material defaultMat;       // default material
  
  double jeps;
  double invjScale, twojeps; // precomputed values for speed
  double jScale;             // scale of quaternion fractal
  SbVec3f jOrig;             // translation of quaternion fractal
  SbMatrix jRotate;
  Quat fc;                   // c value in z(n+1) = z(n)^2 + c
  Quat aB, bB, cB;           // basis of rotate quaternion
  SimpleSphere juliaSphere;  // bounding sphere of fractal
  
//   bool f(const Quat &z_, const Quat &c, int n=15, double big=1e5) const {
//     Quat z = z_;
//     for (int i=0; i<n; i++)
//       z = z*z + c;
//     return z.len2()>big*big;
//   }

//   double fd(const Quat &z_, const Quat &c, int n=15, double big=1e5) const {
//     Quat z = z_;
//     for (int i=0; i<n; i++){
//       z = z*z + c;
//       if (z.len2()>big*big) return i/(float)n;
//     }
//     return z.len2()>big*big ? 1.0 : 0.0;
//   }
  

  void initBasis(int seed=1);

  double dist(const Quat &z_) const;
  
  Quat worldVecToQuat(const SbVec3f &x) const {
    SbVec3f tran = (x-jOrig) * invjScale;
    jRotate.multDirMatrix(tran, tran);
    return aB*tran[0] + bB*tran[1] + cB*tran[2];
  }

  Quat worldDirToQuat(const SbVec3f &x) const {
    SbVec3f tran = x * invjScale;   
    jRotate.multDirMatrix(tran, tran);
    Quat dir = aB*tran[0] + bB*tran[1] + cB*tran[2];
    return dir.norm();
  }
  
  double distVec(const SbVec3f &p) const {
    Quat q = worldVecToQuat(p);
    return dist(q)*jScale;
  }
  
public:

  Julia(int seed=1) {
    jeps = 0.0005;
    jScale = 40.0;
    jRotate.setRotate(SbRotation(SbVec3f(0,1,0), -.78));
    jRotate = jRotate.inverse();

    // precompute values
    invjScale = 1/jScale;
    twojeps = jeps*2;
    
    jOrig = SbVec3f(50, 42.5, 85);
    juliaSphere = SimpleSphere(1.5*jScale, jOrig);    // bounding sphere
    fc = Quat(0.0601525,0.742382,0.0912162,0.00804881);
    initBasis(seed);
  }
  
  virtual PrimType getType() const { return PRIMTYPE_JULIA; }

  bool intersectRay(Ray &r, double t0=1e-5, double t1=1e8);

  SbVec3f getMin() const {
    float r = juliaSphere.rad;
    return juliaSphere.pos - SbVec3f(r,r,r);
  }
  SbVec3f getMax() const {
    float r = juliaSphere.rad;
    return juliaSphere.pos + SbVec3f(r,r,r);
  }

  Material* getMat() { return &defaultMat; }
  const Material* getMat() const { return &defaultMat; }

  SbColor getIntDiffColor(const SbVec3f &bc) const { return getMat()->diffuse; }
  SbVec3f getNormal(const SbVec3f &bc)       const {

    // bc contains position
    const double e = jeps*jScale*0.25;
    double dd = distVec(bc); // this could be reused from before..
    double dx = distVec(bc + SbVec3f(e,0,0));
    double dy = distVec(bc + SbVec3f(0,e,0));
    double dz = distVec(bc + SbVec3f(0,0,e));
    SbVec3f n(dx-dd, dy-dd, dz-dd);
    n.normalize();
    return n;
  }

  static Material* getDefaultMat() { return &defaultMat; }
};


#endif
