#ifndef SBVEC3D_H
#define SBVEC3D_H

#include <ostream>
#include <Inventor/SbBasic.h>
#include <Inventor/SbLinear.h>

using namespace std;

#define DELTA 1e-6

class SbVec3d {
  public:
    // Default constructor
    SbVec3d()						{ }

  // glue
  SbVec3d(const SbVec3f &x){
    vec[0] = x[0];
    vec[1] = x[1];
    vec[2] = x[2];
  }
  SbVec3f f() const { return SbVec3f(vec[0], vec[1], vec[2]); }

    // Constructor given an array of 3 components
    SbVec3d(const double v[3])
	 { vec[0] = v[0]; vec[1] = v[1]; vec[2] = v[2]; }

    // Constructor given 3 individual components
    SbVec3d(double x, double y, double z)
	 { vec[0] = x; vec[1] = y; vec[2] = z; }

    // Constructor given 3 planes
    SbVec3d(SbPlane &p0, SbPlane &p1, SbPlane &p2);

// C-api: begin
    // Returns right-handed cross product of vector and another vector
    SbVec3d	cross(const SbVec3d &v) const;
// C-api: end

    // Returns dot (inner) product of vector and another vector
    double	dot(const SbVec3d &v) const;
// C-api.h: #define SbV3fDot(_v0, _v1)
// C-api.h:     ((_v0).vec[0] * (_v1).vec[0] + (_v0).vec[1] * (_v1).vec[1] +
// C-api.h:      (_v0).vec[2] * (_v1).vec[2])

    // Returns pointer to array of 3 components
    const double	*getValue() const			{ return vec; }
// C-api.h: #define SbV3fGetXYZ(_xyz, _src)
// C-api.h:     (((_xyz)[0] = (_src).vec[0]), ((_xyz)[1] = (_src).vec[1]),
// C-api.h:	 ((_xyz)[2] = (_src).vec[2]))

    // Returns 3 individual components
    void	getValue(double &x, double &y, double &z) const;
// C-api.h: #define SbV3fGetX_Y_Z(_x, _y, _z, _src)
// C-api.h:     (((_x) = (_src).vec[0]), ((_y) = (_src).vec[1]), ((_z) = (_src).vec[2]))

    // Returns geometric length of vector
    double	length() const;
// C-api.h: #define SbV3fLen(_v)
// C-api.h:     (sqrtf(SbV3fDot((_v), (_v))))

// C-api: begin
    // Changes vector to be unit length
    // C-api: name=norm
    double	normalize();
// C-api: end

    // Negates each component of vector in place
    void	negate();
// C-api.h: #define SbV3fNegate(_v)
// C-api.h: 	SbV3fMultBy(_v, -1.0)

    // Sets value of vector from array of 3 components
    SbVec3d &	setValue(const double v[3])
	 { vec[0] = v[0]; vec[1] = v[1]; vec[2] = v[2]; return *this; }
// C-api.h: #define SbV3fSetXYZ(_dest, _src)
// C-api.h:     (((_dest).vec[0] = (_src)[0]), ((_dest).vec[1] = (_src)[1]),
// C-api.h:      ((_dest).vec[2] = (_src)[2]))

    // Sets value of vector from 3 individual components
    SbVec3d &	setValue(double x, double y, double z)
	 { vec[0] = x; vec[1] = y; vec[2] = z; return *this; }
// C-api.h: #define SbV3fSetX_Y_Z(_dest, _x, _y, _z)
// C-api.h:     (((_dest).vec[0] = (_x)), ((_dest).vec[1] = (_y)),
// C-api.h:      ((_dest).vec[2] = (_z)))

    // Sets value of vector to be convex combination of 3 other
    // vectors, using barycentic coordinates
    SbVec3d &	setValue(const SbVec3d &barycentic,
		const SbVec3d &v0, const SbVec3d &v1, const SbVec3d &v2);

    // Accesses indexed component of vector
    double &	  operator [](int i) 		{ return (vec[i]); }
    const double & operator [](int i) const	{ return (vec[i]); }

    // Component-wise scalar multiplication and division operators
    SbVec3d &	operator *=(double d);
// C-api.h: #define SbV3fMultBy(_v, _s)
// C-api.h:     (((_v).vec[0] *= (_s)), ((_v).vec[1] *= (_s)), ((_v).vec[2] *= (_s)))

    SbVec3d &	operator /=(double d)
	{ return *this *= (1.0 / d); }
// C-api.h: #define SbV3fDivBy(_v, _s)
// C-api.h:     (((_v).vec[0] /= (_s)), ((_v).vec[1] /= (_s)), ((_v).vec[2] /= (_s)))

    // Component-wise vector addition and subtraction operators
    SbVec3d &	operator +=(SbVec3d v);
    SbVec3d &	operator -=(SbVec3d v);

    // Nondestructive unary negation - returns a new vector
    SbVec3d	operator -() const;

    // Component-wise binary scalar multiplication and division operators
    friend SbVec3d	operator *(const SbVec3d &v, double d);
    friend SbVec3d	operator *(double d, const SbVec3d &v)
	{ return v * d; }
    friend SbVec3d	operator /(const SbVec3d &v, double d)
	{ return v * (1.0 / d); }

    // Component-wise binary vector addition and subtraction operators
    friend SbVec3d	operator +(const SbVec3d &v1, const SbVec3d &v2);
// C-api.h: #define SbV3fAdd(_dest, _src1, _src2)
// C-api.h:     (((_dest).vec[0] = (_src1).vec[0] + (_src2).vec[0]),
// C-api.h:      ((_dest).vec[1] = (_src1).vec[1] + (_src2).vec[1]),
// C-api.h:      ((_dest).vec[2] = (_src1).vec[2] + (_src2).vec[2]))

    friend SbVec3d	operator -(const SbVec3d &v1, const SbVec3d &v2);
// C-api.h: #define SbV3fSub(_dest, _src1, _src2)
// C-api.h:     (((_dest).vec[0] = (_src1).vec[0] - (_src2).vec[0]),
// C-api.h:      ((_dest).vec[1] = (_src1).vec[1] - (_src2).vec[1]),
// C-api.h:      ((_dest).vec[2] = (_src1).vec[2] - (_src2).vec[2]))

// C-api: begin
    // Equality comparison operator
    friend int		operator ==(const SbVec3d &v1, const SbVec3d &v2);
    friend int		operator !=(const SbVec3d &v1, const SbVec3d &v2)
	{ return !(v1 == v2); }

    // Equality comparison within given tolerance - the square of the
    // length of the maximum distance between the two vectors
    SbBool		equals(const SbVec3d v, double tolerance) const;

    // Returns principal axis that is closest (based on maximum dot
    // product) to this vector
    SbVec3d		getClosestAxis() const;

  protected:
    double	vec[3];		// Storage for vector components
};

ostream& operator<< (ostream& os, const SbVec3d& x);

#endif
