#include <string>
#include "SkyConstants.h"
#include "Sky.h"
#include "Spectrum.h"
#include "DataSet.h"
#include "debug.h"

using namespace std;

//     Y N
//     |/
//  W--x--X(E)
//    /|
//   Z(S)

// Implementation of "A Practical Analytic Model for Daylight"
// by  A. J. Preetham, Peter Shirley, Brian Smits
// http://www.cs.utah.edu/vissim/papers/sunsky/

// by Kevin Beason

Sky::Sky(){
  sunBasisU = SbVec3f(1,0,0);
  sunBasisV = SbVec3f(0,0,-1);
  sunBasisW = SbVec3f(0,1,0);
  sunRadius = sunRadiusAngle = sunDist = sunOmega = 0;
  turbidity = thetaSolar = thetaSolarSafe = 0;
  sunRadiance.setValue(0,0,0);
  AY=BY=CY=DY=EY=Ax=Bx=Cx=Dx=Ex=Ay=By=Cy=Dy=Ey=0;
  Yz=xz=yz=0;
}

Sky::~Sky(){
}

void Sky::init(GlobalVars *gv_, float stdTime) {

  // set GlobalVars pointer
  gv = gv_;

//  float stdTime = stdTimeStr ? atof(stdTimeStr) : 12+5;
//  float stdTime = 12+3;     // hours, decimal. 3pm
//  float stdTime = 12+5;     // hours, decimal. 6pm
//  float julianDate = 322;   // 11/18/07  see: `date +%j`
  float julianDate = 66;   // 3/6/08  see: `date +%j`
  float stdMeridianPST = -8.*15 * M_PI/180;  // radians
  float stdMeridian = stdMeridianPST;
  float latitude = 34.017 * M_PI/180;  // los angeles
  float longitude = -118.41 * M_PI/180; // http://www.gorissen.info/Pierre/maps/googleMapLocationv3.php

  // in decimal hours
  float solarTime = stdTime +
    0.170 * sin(4.*M_PI*(julianDate-80)/373.) -
    0.129 * sin(2.*M_PI*(julianDate-8)/355.) +
    12. * (stdMeridian-longitude)/M_PI;

  float solarDeclination = 0.4093 * sin(2.*M_PI*(julianDate-81)/368.); // radians "delta"
  float tmp = M_PI*solarTime/12.;

  // solar angle from zenith in radians
  // [0, pi/2]
  thetaSolar = M_PI/2. - asin(sin(latitude)*sin(solarDeclination) -
                              cos(latitude)*cos(solarDeclination) *
                              cos(tmp));

  // solar azimuth in radians, west from South
  // [0, 2pi]
  float opp = -cos(solarDeclination)*sin(tmp);
  float adj = cos(latitude)*sin(solarDeclination) -
              sin(latitude)*cos(solarDeclination)*cos(tmp);
  float phiSolar = atan2(opp, adj);

  //phiSolar=thetaSolar=0;

  // sun distance, in meters. perihelion/aphelion
  // distances from http://www.physicalgeography.net/fundamentals/6h.html
  float sunDistJul4 = 152.1e9; // far
  float sunDistJan3 = 147.3e9; // near

  // lerp... not accurate but whateva
  float yearFrac = (julianDate-3+365%365)/365.;
  if (yearFrac>.5) yearFrac = 1-yearFrac;
  sunDist = (1-yearFrac)*sunDistJan3 + yearFrac*sunDistJul4;

  sunRadius = 695500e3; // in meters (thanks google)
  //sunDist = 1.5e11;
  sunRadiusAngle = atan2(sunRadius, sunDist);  //sunSize/sunDist;

  // compute solid angle
  sunOmega = M_PI*sunRadius*sunRadius / (4*M_PI*sunDist*sunDist) * 4*M_PI;
  
  // orthonormal basis (W) pointed towards sun
  // phi = 0 = noon = south = z
  // phi = pi/2 = evening = west = -x
  // up = y
  cerr << "phiSolar = " << phiSolar << endl;
  float phi = phiSolar; //1.5*M_PI-phiSolar;
  cerr << "phi = " << phi << endl;
  sunBasisW.setValue(sin(thetaSolar)*-sin(phi), // FIXME? Is the phi part right?
		     cos(thetaSolar),
		     sin(thetaSolar)*cos(phi));
  sunBasisV.setValue(sunBasisW[1], sunBasisW[2], sunBasisW[0]);
  sunBasisV -= sunBasisV.dot(sunBasisW)*sunBasisW;
  sunBasisV.normalize();
  sunBasisU = sunBasisV.cross(sunBasisW);

  cerr << "Sun W = " << sunBasisW << endl;

  Spectrum sunSpec(solAmplitudes, 38, 380, 750);
  sunSpec.scaleSpec(10); // values were off by 10
  sunSpec.scaleSpec(10); // convert to W/(m^2 sr nm)
  //sunSpec.dump("utah.csv");
  
  // modulate color based on atmosphere absorbtion

  // clamp to pi/2
  float thetaSolarSafe = thetaSolar;
  if (thetaSolarSafe>M_PI/2)
    thetaSolarSafe = M_PI/2;
  
  // reverse at pi/2
//   if (thetaSolarSafe>M_PI/2)
//     thetaSolarSafe = M_PI-thetaSolarSafe;
  
  cerr << "thetaSolar = " << thetaSolar << endl;
  cerr << "thetaSolarSafe = " << thetaSolarSafe << endl;

  turbidity = 2;
  float beta = 0.04608*turbidity-0.04586;
  float alpha = 1.3;
  float l = 0.35;
  float w = 2;
  float m = 1./(cos(thetaSolarSafe)+0.15*
                pow(93.885-thetaSolarSafe*180/M_PI, -1.253));

  // need wl, k_o, k_g, k_wa
  SpectralCurveIrreg k_oFunc(k_oWavelengths, k_oAmplitudes, 64);
  SpectralCurveIrreg k_gFunc(k_gWavelengths, k_gAmplitudes, 4);
  SpectralCurveIrreg k_waFunc(k_waWavelengths, k_waAmplitudes, 13);

  int num = Spectrum::getNumWavelens();
  float max = Spectrum::getMaxWavelen();
  float min = Spectrum::getMinWavelen();
  cerr << "m=" << m << endl;
  cerr << "num=" << num << " max=" << max << " min=" << min << endl;
  for (int i=0; i<num; i++){
    float wl = (max-min) * i/(float)(num-1) + min;

    // apparently lambda should be in um
    float wlum = wl/1000;
    //cerr << "wl=" << wl << endl;

    float k_o = k_oFunc(wl);
    float k_g = k_gFunc(wl);
    float k_wa = k_waFunc(wl);
    
    float t_air = exp(-m * 0.008735 * powf(wlum, -4.08));
    float t_ang = exp(-m * beta * pow(wlum, -alpha));
    float t_ozo = exp(-m * k_o * l);
    float t_gas = exp(-1.41 * k_g * m / pow(1 + 118.93 * k_g * m, 0.45));
    float t_h2o = exp(-0.2385 * k_wa * w * m /
                      pow(1 + 20.07 * k_wa * w * m, 0.45));

    float t_tot;
    t_tot = t_air*t_ang*t_ozo*t_gas*t_h2o;
//  t_tot = t_air;
//    cerr << "t_tot[" << i << "=" << wl << "] = " << t_tot << endl;
//     cerr << "[" << i << "=" << wl << "] " <<
//       "k_o="  << k_o << " " <<
//       "k_wa=" << k_wa << " " <<
//       "k_g="  << k_g << endl;
      
      
    sunSpec.scaleAmp(i, t_tot);
  }
  
  sunRadiance = sunSpec.getRGB();

  cerr << "Sun radiance = " << sunRadiance << endl;

  
  // constants
  // these vales taken from source rather than paper
  // http://www.cs.utah.edu/vissim/papers/sunsky/code
  float T=turbidity;
  AY =  0.17872*T + -1.46303;
  BY = -0.35540*T +  0.42749;
  CY = -0.02266*T +  5.32505;
  DY =  0.12064*T + -2.57705;
  EY = -0.06696*T +  0.37027;

  Ax = -0.01925*T + -0.25922;
  Bx = -0.06651*T +  0.00081;
  Cx = -0.00041*T +  0.21247;
  Dx = -0.06409*T + -0.89887;
  Ex = -0.00325*T +  0.04517;

  Ay = -0.01669*T + -0.26078;
  By = -0.09495*T +  0.00921;
  Cy = -0.00792*T +  0.21023;
  Dy = -0.04405*T + -1.65369;
  Ey = -0.01092*T +  0.05291;

  float t = thetaSolar;
  float t2 = t*t, t3=t*t*t;
  float chi = (4./9.-T/120.)*(M_PI-2*t);
  Yz = (4.0453*T-4.9710)*tan(chi) - 0.2155*T + 2.4192;
  xz =
    ( 0.00165*t3 + -0.00374*t2 +  0.00208*t + 0.0    )*T*T +
    (-0.02902*t3 +  0.06377*t2 + -0.03202*t + 0.00394)*T   +
    ( 0.11693*t3 + -0.21196*t2 +  0.06052*t + 0.25885);
  yz =
    ( 0.00275*t3 + -0.00610*t2 +  0.00316*t + 0.0    )*T*T +
    (-0.04214*t3 +  0.08970*t2 + -0.04153*t + 0.00515)*T   +
    ( 0.15346*t3 + -0.26756*t2 +  0.06669*t + 0.26688);

  // convert Kcd/m^2 to cd/m^2 to W/m^2sr
  Yz *= 1000/683.;

  // render sky view to environment map
  cerr << "Precomputing sky HDR map... ";
  SbVec3f ndc;
  int dim[2] = {512, 512};
  float *texture = new float[3 * dim[0]*dim[1]];

  for (int j=0; j<dim[1]; j++){
    for (int i=0; i<dim[0]; i++){

      ndc[0] =  (((double)i + 0.5) / dim[0] - 0.5)*2.0;
      ndc[1] =  (((double)j + 0.5) / dim[1] - 0.5)*2.0;
      ndc[2] = 0;

      int k = (j*dim[0]+i)*3;
      
      float theta = ndc.length() * M_PI; //  /2;
      
      if (theta>M_PI_2) {
        texture[k+0]=texture[k+1]=texture[k+2] = 0;
        continue;
      }
        
      float phi = atan2(ndc[1], ndc[0]);
      SbVec3f dir = SbVec3f(
        sin(theta)*cos(phi),
        cos(theta),
        sin(theta)*-sin(phi)
        );

      SbColor col = getRadiance(dir, false);
      texture[k+0] = col[0];
      texture[k+1] = col[1];
      texture[k+2] = col[2];
    }
  }

  cerr << "done.\n";

  // calculate rotation
  float angle = gv->envMapRotation*M_PI/180.0; // convert degrees to radians
  SbRotation userSpin(SbVec3f(0.0, 1.0, 0.0), angle); // user angle about Y axis
  SbRotation skyCorrection(SbVec3f(1,0,0),-M_PI/2); // 90 degree fix about X
  SbRotation skyRot = skyCorrection*userSpin;
  
  // initialize texture subclass
  sky.init(gv, texture, dim[0], dim[1], &skyRot);
  //string file = "envmap.hdr";
  //sky.writeDataHdr(file);

  // now use rotation
  setRotateSky(userSpin);
  
  delete [] texture;
}

SbColor Sky::impSample(Ray &r, SbVec3f &n, SbVec3f &ng, int simple) {

  int samples = 4;
  if (samples>gv->lightSamples)
    samples = gv->lightSamples;
  
  int sqrt_samples = (int)sqrt((float)samples); // this should be precomputed

  // use one sample in final gather
  if (simple)
   sqrt_samples = 1;
  
  if (sqrt_samples<1) sqrt_samples = 1;
  int numVis = 0;
  SbColor color(0,0,0);

  float sun_w = M_PI*sunRadius*sunRadius / (4*M_PI*sunDist*sunDist) *4*M_PI;
  float density = 1/sun_w;
  float densityInv = 1/density;
  
  for (int sampy=0; sampy<sqrt_samples; sampy++){
    for (int sampx=0; sampx<sqrt_samples; sampx++){

      // jittered
      double eps1 = (sampx + r.ctx->mydrand48()) / sqrt_samples;
      double eps2 = (sampy + r.ctx->mydrand48()) / sqrt_samples;

      // sample sun disc uniformly
      // Shirley's RRT 1e pg. 123
      double cosAlpha = 1+eps1*(cos(sunRadiusAngle)-1);
      double sinAlpha = sqrt(1-cosAlpha*cosAlpha);
      double phi = 2*M_PI*eps2;
      SbVec3f dir = 
 	sinAlpha*cos(phi)*sunBasisU +
 	sinAlpha*sin(phi)*sunBasisV +
 	cosAlpha*sunBasisW;

      // rotate
      textureToWorldRot.multDirMatrix(dir, dir);

      // dont illuminate lights below horizon
      if (ng.dot(dir) <= 0.0)
	continue;
      
      // don't illuminate area that think they are facing away
      float dot1 = n.dot(dir);
      if (dot1<=0.0)
	continue;      
		  
      const float t_eps = 0.002;
      Ray lr = Ray(r.x+dir*t_eps, dir, r.ctx);
      lr.ignore(r.prim);
      if (gv->pane->scenePrim->intersectRay(lr)){
	continue; // hits something (and not background)
      }

      float value = dot1;

      // ray marching absorbtion
      if (gv->stepSize>0){
	float tau = 0.0;
	float sdt = gv->stepSize;
	float sigma_t = 0.0;
	SbVec3f lx = r.x+gv->pane->getMaxDist()*dir;
	tau = gv->pane->dataSet->bresenSum(r.x, lx);
	tau = gv->meters(tau);
	value *= exp(-tau);
      }
	  
      // compute color
      // shirley p.120 eq. 15.5
      color += SbColor(r.diffCol[R] * M_1_PI * sunRadiance[R] * value,
		       r.diffCol[G] * M_1_PI * sunRadiance[G] * value,
		       r.diffCol[B] * M_1_PI * sunRadiance[B] * value) ;

    }
  }

  // cerr << "sunRadiance = " << sunRadiance << endl;
  // cerr << "sunDist = " << sunDist << endl;
  // cerr << "sunW = " << sun_w << endl;
  color *= densityInv / (sqrt_samples*sqrt_samples);

  color += sky.impSample(r, n, ng, simple);

  return color;
}

float Sky::getSample(double eps1, double eps2,
                     SbVec3f &dir, SbColor &c){

  if (eps2>=0.5){  // sun
    textureToWorldRot.multDirMatrix(sunBasisW, dir);
    c = sunRadiance;
    return 1/sunOmega*0.5;
  } else {        // sky
    return sky.getSample(eps1, eps2*2.0, dir, c)*0.5;
  }
}


namespace {

  float Perez(float theta, float gamma, float zenVal, float thetaSun,
              float A, float B, float C, float D, float E){
    double cosgamma = cos(gamma);
    double costheta = cos(theta);
    double cosThetaSun = cos(thetaSun);

    double BoverCT = B/costheta;
    if (BoverCT > 20)
      BoverCT = 20;

    double num =
      (1. + A*exp(BoverCT)) *
      (1. + C*exp(D*gamma) + E*cosgamma*cosgamma);

    double den =
      (1. + A*exp(B)) *
      (1. + C*exp(D*thetaSun) + E*cosThetaSun*cosThetaSun);

    double rat = num/den;
    if (rat>1e2)
      rat = 1e2;
    else if (rat<-1e2)
      rat = -1e2;

    return rat * zenVal;
    
  }
}

SbColor Sky::getRadiance(SbVec3f &dir, bool includeSun){

  SbColor color(0,0,0);

  // rotate
  worldToTextureRot.multDirMatrix(dir, dir);
  
  float theta = acosf(dir.dot(SbVec3f(0,1,0)));

  float dirDotSunW = dir.dot(sunBasisW);
  // clip to [-1,1]
  if (dirDotSunW > 1) dirDotSunW = 1;
  else if (dirDotSunW < -1) dirDotSunW = -1;
  float gamma = acosf(dirDotSunW);

  if (includeSun) {        // direct ray

    if (gamma<sunRadiusAngle)           // include sun
      color += sunRadiance;

    if (theta>M_PI_2)                   // work below horizon
      theta = M_PI_2-(theta-M_PI_2);
  }
  else if (dir[1]<=0)      // for precomputing sky
    return SbColor(0,0,0); // nothing below horizon

  static float Ym[] = { 1e9,-1e9 };
  static float xm[] = { 1e9,-1e9 };
  static float ym[] = { 1e9,-1e9 };

  float Y =  Perez(theta,gamma,Yz,thetaSolar,AY,BY,CY,DY,EY);
  if (Y<0)
    return SbColor(0,0,0);

  bool say = false;
  if (Y<Ym[0]){
    Ym[0] = Y;
    say = true;
  } else if (Y>Ym[1]) {
    Ym[1] = Y;
    say = true;
  }
  
  float x = Perez(theta,gamma,xz,thetaSolar,Ax,Bx,Cx,Dx,Ex);

  if (x<xm[0]){
    xm[0] = x;
    say = true;
  } else if (x>xm[1]) {
    xm[1] = x;
    say = true;
  }
  
  float y = Perez(theta,gamma,yz,thetaSolar,Ay,By,Cy,Dy,Ey);

  if (y<ym[0]){
    ym[0] = y;
    say = true;
  } else if (y>ym[1]) {
    ym[1] = y;
    say = true;
  }

  color += Spectrum::YxyToRgb(Y,x,y);
  
  return color;
}

SbColor Sky::getEmissiveFlux(float receiverArea){
  
  // L = Phi/(Area*cosTheta*Omega) -> Phi = L*Area*cosTheta*Omega
  //
  // (where Area = pi*worldRad^2, cosTheta = 1, Omega={sunW,4*pi} )
  //
  // EmisSun = sunRadiance * AreaDisc * 1 * sunOmega    (theta is unknown)
  // EmisSky = sky.Lavg    * AreaDisc * 1 * 4*pi
  //
  // EmisTot = EmisSun+EmisSky

  // L = Phi/(Area*cosTheta*Omega) -> Phi = L*Area*cosTheta*Omega
  SbVec3f sunFlux = sunRadiance * receiverArea * sunOmega;
  SbVec3f skyFlux = sky.getEmissiveFlux(receiverArea);
  return sunFlux+skyFlux;
}
