#ifndef SPHERE_H
#define SPHERE_H

#include "Primitive.h"
#include "Material.h"
#include "Texture.h"

class SphereSet {
public:
  vector<SbColor>    colors;
  vector<Material *> mtls;
  vector<SbVec3f>    verts;     // optional
  vector<float>      radii;     // optional
  vector<Texture *>  textures;  // optional

  int add(const SbVec3f &vert, float radius, const SbColor &color,
          Material *mat, Texture *texture) {
    int idx = verts.size();
    colors.push_back(color);
    mtls.push_back(mat);
    verts.push_back(vert);
    radii.push_back(radius);
    textures.push_back(texture);
    return idx;
  }
  int add(const SbVec3f &vert, float radius, const SbColor &color,
          Material *mat) {
    int idx = verts.size();
    colors.push_back(color);
    mtls.push_back(mat);
    verts.push_back(vert);
    radii.push_back(radius);
    return idx;
  }
  int add(const SbColor &color, Material *mat, Texture *texture) {
    int idx = verts.size();
    colors.push_back(color);
    mtls.push_back(mat);
    textures.push_back(texture);
    return idx;
  }
  int add(const SbColor &color, Material *mat) {
    int idx = verts.size();
    colors.push_back(color);
    mtls.push_back(mat);
    return idx;
  }
};

class Sphere : public Primitive {
    
private:

  // verts, radii, colors, mtls
  static SphereSet defaultSphereSet;
  static SphereSet defaultTextureSphereSet; // for spheres with textures

  // colors, mtls
  static SphereSet unitSphereSet;           // for instanced spheres
  static SphereSet unitTextureSphereSet;    // for unit spheres with textures
                                   

  static Material defaultMat;
  static SbVec3f origin;
  
  SphereSet *sphereSet;
  int n;

public:

  Sphere(const SbVec3f &pos=SbVec3f(0,0,0), float radius=1,
         const SbVec3f &color=SbColor(.8,.8,.8),
         Material *mat = &defaultMat, Texture *texture=NULL) {

    if (pos[0]!=0 || pos[1]!=0 || pos[2]!= 0 || radius != 1){
      if (texture){
        sphereSet = &defaultTextureSphereSet;
        n = sphereSet->add(pos, radius, color, mat, texture);
      }
      else {
        sphereSet = &defaultSphereSet;
        n = sphereSet->add(pos, radius, color, mat);
      }
    }
    else {
      if (texture){
        sphereSet = &unitTextureSphereSet;
        n = sphereSet->add(color, mat, texture);
      }
      else {
        sphereSet = &unitSphereSet;
        n = sphereSet->add(color, mat);
      }
    }
  }

  virtual PrimType getType() const { return PRIMTYPE_SPHERE; }

  const SbVec3f& pos() const {
    return hasPos() ? sphereSet->verts[n] : origin;
  }
  const float rad() const {
    return hasRadius() ? sphereSet->radii[n] : 1;
  }
  const SbColor& col() const { return sphereSet->colors[n]; }
  const Texture* texture() const {
    return hasTexture() ? sphereSet->textures[n] : NULL;
  }

  bool hasPos() const { return !sphereSet->verts.empty(); }
  bool hasRadius() const { return !sphereSet->radii.empty(); }
  bool hasTexture() const { return !sphereSet->textures.empty(); }
  
  bool intersectRay(Ray &r, double t0=1e-5, double t1=1e8);

  SbVec3f getMin() const {
    float r = rad();
    return pos() - SbVec3f(r,r,r);
  }
  SbVec3f getMax() const {
    float r = rad();
    return pos() + SbVec3f(r,r,r);
  }

  Material* getMat() { return sphereSet->mtls[n]; }
  const Material* getMat() const { return sphereSet->mtls[n]; }
  SbColor getIntDiffColor(const SbVec3f &bc) const;
  SbVec3f getNormal(const SbVec3f &bc) const {
    SbVec3f n = bc-pos();
    n.normalize();
    return n;
  }
  SbVec3f getIntNormal(const SbVec3f &bc) const;
  bool    getDerivs(      const SbVec3f &bc, PrimDerivs &pDer) const;

  void getSTfromBC(const SbVec3f &bc, float &s, float &t) const {
    SbVec3f rad = bc-pos();
    rad.normalize();
    s = atan2f(rad[0], rad[2])*float(0.5/M_PI) + 0.5f;
    t = acosf(-rad[1])*float(1./M_PI);

    // transform by texture matrix..
    const Texture *text = texture();
    SbMatrix m;
    m.makeIdentity();
    m[0][0] = text->stex[0];
    m[1][0] = text->stex[1];
    m[3][0] = text->stex[2];
    m[0][1] = text->ttex[0];
    m[1][1] = text->ttex[1];
    m[3][1] = text->ttex[2];
    SbVec3f coord;
    m.multVecMatrix(SbVec3f(s,t,0), coord);
    s = coord[0];
    t = coord[1];
  }
  
  static Material* getDefaultMat() { return &defaultMat; }
};

#endif
