////////////////////////////////////////////////////////////////////////////
//
// File: intersect_ray_tri.cxx
// Description: Pane class (Core Ray Tracing Engine) header
// Created: 2003/07/17 09:33:17
// Author: Kevin Beason <beason@cs.fsu.edu>
//
// Copyright (c) 2003-2004 Kevin Beason
// 
// ray-triangle intersection subroutine
// for the Pane ray tracer.
// For more information on Pane, visit:
// http://www.csit.fsu.edu/~beason/pane/
// or see the accompanying documentation
// in the distribution.
//
// Quickly determines whether a given
// ray intersects a given triangle.
//
// This code is from
// http://www.ce.chalmers.se/staff/tomasm/raytri/raytri.c
// with slight modifications.
//
// $Id: intersect_ray_tri.cxx 1158 2007-05-11 06:18:18Z beason $
//
////////////////////////////////////////////////////////////////////////////

#include "intersect_ray_tri.h"

#define EPSILON 0.000000000001
#define CROSS(dest,v1,v2) \
              dest[0]=v1[1]*v2[2]-v1[2]*v2[1]; \
              dest[1]=v1[2]*v2[0]-v1[0]*v2[2]; \
              dest[2]=v1[0]*v2[1]-v1[1]*v2[0];
#define DOT(v1,v2) (v1[0]*v2[0]+v1[1]*v2[1]+v1[2]*v2[2])
#define SUB(dest,v1,v2) \
            dest[0]=v1[0]-v2[0]; \
            dest[1]=v1[1]-v2[1]; \
            dest[2]=v1[2]-v2[2]; 


// code rewritten to do tests on the sign of the determinant 
// the division is at the end in the code                    
int intersect_ray_tri(const SbVec3f &orig, const SbVec3f &dir,
		      const SbVec3f &vert0, const SbVec3f &vert1, const SbVec3f &vert2,
		      double *t, double *u, double *v)
{

  // the precision of the types below can impact speed and accuracy
  // greatly. tweak if you have problems with cracks (or don't).
  double edge1[3], edge2[3];
  double tvec[3];
  double pvec[3];
  double qvec[3];
  double det;
  double inv_det;

  // find vectors for two edges sharing vert0 
  SUB(edge1, vert1, vert0);
  SUB(edge2, vert2, vert0);

  // begin calculating determinant - also used to calculate U parameter
  CROSS(pvec, dir, edge2);

  // if determinant is near zero, ray lies in plane of triangle
  det = DOT(edge1, pvec);

  if (det > EPSILON){
     
    // calculate distance from vert0 to ray origin 
    SUB(tvec, orig, vert0);
      
    // calculate U parameter and test bounds
    *u = DOT(tvec, pvec);
    if (*u < 0.0 || *u > det)
      return 0;
     
    // prepare to test V parameter
    CROSS(qvec, tvec, edge1);
      
    // calculate V parameter and test bounds
    *v = DOT(dir, qvec);
    if (*v < 0.0 || *u + *v > det)
      return 0;
      
  }
  else if(det < -EPSILON){
     
    // calculate distance from vert0 to ray origin
    SUB(tvec, orig, vert0);
      
    // calculate U parameter and test bounds
    *u = DOT(tvec, pvec);
    if (*u > 0.0 || *u < det)
      return 0;
      
    // prepare to test V parameter
    CROSS(qvec, tvec, edge1);
      
    // calculate V parameter and test bounds
    *v = DOT(dir, qvec) ;
    if (*v > 0.0 || *u + *v < det)
      return 0;
  }
  else return 0;  // ray is parallell to the plane of the triangle

  inv_det = 1.0 / det;
   
  // calculate t, ray intersects triangle
  *t = DOT(edge2, qvec) * inv_det;
  (*u) *= inv_det;
  (*v) *= inv_det;

  return 1;
}
